
"""
decocare - a pure python module for serial communication with insulin pumps.

This module implements the communication protocol necessary to audit
or command MM insulin pumps.  Tested with 515, and 522.

See:

.. hlist::
  * https://github.com/bewest/decoding-carelink
  * https://github.com/bewest/insulaudit
  * https://github.com/bewest/insulaudit-ssh-tools
  * https://github.com/bewest/diabetes
  * https://github.com/medevice-users/diabetes#projects


"""

